/* 
 * CSE142, Spring 2000
 * Homework 4
 * Name:
 * SID:
 * Section:
 *
 *    Evil Evil Monster Monster  -- an interactive game from HHGG
 *
 */

#include <stdio.h>
#include <stdlib.h> /* needed for rand() */
#include <math.h>   /* we use abs() -- absolute value -- to help with distance() */
#include <time.h>   /* we use time() to setup the random number generator */

#include "gp142.h"  /* needed for all the drawing functions */



/*****************************************************************
 * CONSTANTS
 *****************************************************************/

/* Note - in C you can define a constant in terms of other, earlier constants, as
   you will see below */


#define FALSE 0  /* boolean values */
#define TRUE  1

/* 
 * For many of the GP142 functions, a line width of
 * zero means to fill the shape being drawn.  The
 * constant FILL is used in function calls when a solid
 * shape is desired.
 */
#define FILL 0 

#define POINT_SIZE 12  /* default font size for things written using GP142_printfXY */


#define PLAYER_WIDTH  40  /* width (diameter) of the player and monster icons */
#define PLAYER_SPEED_CHANGE 5  /* how much the player's speed changes with 1 key press */
#define STARTING_HEALTH 50   /* how much health the player starts with */

#define MONSTER_SPEED 8     /* default speed of monster */
#define MONSTER_COLOR RED   /* default monster color */

#define BUTTON_HEIGHT 25    /* height of a button appearing at the top of the screen */
#define BUTTON_WIDTH  100   /* width of such a button */


/* The following give the bounds of the screen.  No object should appear outside these
   bounds.   */
#define X_LEFT_BOUND    (-GP142_XMAX + PLAYER_WIDTH/2)  
#define X_RIGHT_BOUND   (GP142_XMAX - PLAYER_WIDTH/2)   
#define Y_TOP_BOUND     (GP142_YMAX - BUTTON_HEIGHT - PLAYER_WIDTH/2)
#define Y_BOTTOM_BOUND  (-GP142_YMAX + PLAYER_WIDTH/2)


#define BUTTON_BOTTOM (Y_TOP_BOUND + PLAYER_WIDTH/2) /* The bottom of the buttons starts here */

#define PELLET_WIDTH (PLAYER_WIDTH * 3.0/5.0) /* diameter of a pellet */                                										

#define NUM_PELLET      5  /* how many pellets should appear on the screen */

#define NON_PELLET      0  /* a pellet becomes this type after being eaten -- NON_PELLET
							are not drawn */
/* These are the differnet types of pellets */
#define POINT_PELLET    1  /* just increase score default value */
#define BONUS_PELLET    2  /* gives player twice the normal points */
#define CLEARALL_PELLET 3  /* erases all pellets on the screen and creates NUM_PELLET new ones */
#define HEALTH_PELLET   4  /* increases player's health */

#define PELLET_VALUE    1   /* amount player's score increase by eating 1 pellet-- default value */
#define HEALTH_INCREASE 2   /* amount player's health increases */

/* These are used to determine which button was clicked on */
#define NON_BUTTON      0
#define LEFT_BUTTON     1
#define RIGHT_BUTTON    2

/*****************************************************************
 * FUNCTION PROTOTYPES
 *****************************************************************/


/* See the function definitions for details as to what
 * each function does. Most functions have self-explanatory names.
 */

void draw_player(int player_x, int player_y, int player_color);
void draw_monster(int monster_x, int monster_y, int monster_color);
void draw_pellet(int x, int y, int pellet_type);
void draw_buttons(void);
void draw_game(int player_x, int player_y, int player_health, int player_color, int monster_x, int monster_y, int monster_color, int pellets_x[],
			   int pellets_y[], int pellets_type[], int num_pellets_consumed,int gameOver);


int  find_mouse_event(int mouse_x, int mouse_y);
void handle_mouse_event(int event_type, int *player_x, int *player_y, int *player_color,
						int *player_x_speed, int *player_y_speed, int *player_health,
						int *num_pellets_consumed, int *monster_x, int *monster_y,
						int *monster_speed, int *monster_color, int pellets_x[], int pellets_y[], int pellets_type[],
						int *gameIsPaused, int *gameOver);
void handle_keyboard_event(char key_pressed, int *gameIsPaused, int *player_x_speed, int *player_y_speed);
void handle_monster_attack(int player_x, int player_y, int *player_health, int monster_x,
						   int monster_y, int *gameOver, int *gameIsPaused);
void handle_pellet_consumption(int *player_x, int *player_y, int *player_health, int pellets_x[], int pellets_y[], int pellets_type[],
							   int *num_pellets_consumed, int *monster_speed, int *monster_color);

void update_player_position(int *player_x,int *player_y, int player_x_speed, int player_y_speed);

void pellets_reaction(int *player_x, int *player_y, int *player_health, int pellets_x[], int pellets_y[],
					  int pellets_type[], int pellet_type, int *pellet_value, int *monster_speed, int *monster_color);
void initialize_pellets(int pellets_x[], int pellets_y[], int pellets_type[]);
void pellet_color(int *inner_color, int *outer_color, int pellet_type);
int  new_pellet_type(void);

void enforce_bounds(int *x, int *y);
void new_position(int *x, int *y);
int  distance(int x1, int y1, int x2, int y2);



/*****************************************************************
 * MAIN
 *****************************************************************/

/*
 * Main program contents:
 *   - Major program variables that persist throughout execution
 *   - Event loop that processes GP142 events until the user quits,
 *     and updates variables and redraws the window as needed.
 */

int main(void) {

	/* VARIABLE DECLARATIONS AND INITIALIZATION */

	/* these are used by GP142 events */
	int  quit;             /* does the user wish to quit the game */
	int mouse_x, mouse_y;  /* x and y coordinates of the mouse click */
	char key_pressed;      /* the key pressed by the user */
	

	int mouse_event_type;        /* used to determine which button was pressed */


	/* information about the player */
	int player_x = 0,player_y = 0;   /* player's x and y coordinates */
	int player_x_speed = 0, player_y_speed = 0;  /* player's speed in the x and y directions
													the player's speed in each direction changes
													with key presses */			
	int player_color = YELLOW;   /* player icon color is set to default*/
	int player_health = STARTING_HEALTH;      /* player health is initalized */
	int num_pellets_consumed = 0;/* the player's score */

	/* information about the monster */
	int monster_x = -100, monster_y = 100;  /* monster's coordinates */
	int monster_speed = MONSTER_SPEED;  /* speed monster moves -- note same in both directions */
	int monster_color = MONSTER_COLOR;  /* color of monster */


	/*  pellet information -- an array is used for each of the different major
		pellet attributes.  pellets have an x and y coordinate as well as a type 
		pellet_x[i], pellets_y[i], pellets_type[i] tells everything about the ith pellet */
	int pellets_x[NUM_PELLET];
	int pellets_y[NUM_PELLET];
	int pellets_type[NUM_PELLET];



	/* used to control the game play */
	int gameIsPaused = FALSE;  /* changes each time the p key is pressed */
	int gameOver     = FALSE;  /* is true when player loses all its health */


	/* PROGRAM STARTS HERE */

	srand(time(NULL));  /* initialize the random number generator */

	initialize_pellets(pellets_x,pellets_y,pellets_type); /* initialze the array of pellets */

	/* create the game screen, clear it, and set it up for game play */
	GP142_open();
	GP142_clear();
	GP142_animate(ANI_RUN);
	GP142_logging(LOG_OFF);

	
	/*
	 * Main event loop:
	 * ---- ----- -----
	 * Wait for the next user action, decode it, and call an
	 * appropriate function to handle it.  Repeat until the
	 * user selects quit.
	 */

	quit = FALSE;

	while (!quit) {
		/* get the next GP142 event and act on it */
		switch (GP142_await_event(&mouse_x,&mouse_y,&key_pressed)) {
		case GP142_QUIT:
			quit = TRUE; /* exit the loop and finish the program */
			break;

		/* determine which button was pressed and act on it */
		case GP142_MOUSE:
			mouse_event_type = find_mouse_event(mouse_x,mouse_y);
			handle_mouse_event(mouse_event_type, &player_x, &player_y, &player_color,
						&player_x_speed, &player_y_speed, &player_health,
						&num_pellets_consumed, &monster_x, &monster_y,
						&monster_speed, &monster_color, pellets_x, pellets_y, pellets_type,
						&gameIsPaused, &gameOver);
			break;

		/* while the game is in play, control the character */
		case GP142_KBD:
			if (!gameOver)
				handle_keyboard_event(key_pressed,&gameIsPaused,&player_x_speed,&player_y_speed);
			break;

		/* This is a timer event.  It is an event, like keyboard clicks and mouse presses and
		   which happens roughly 10 times a second.  Each time, everything which needs to be
		   updated, is updated.  Then everything is redrawn. */
		case GP142_PERIODIC:
			if (!gameIsPaused) {
				update_player_position(&player_x,&player_y,player_x_speed,player_y_speed);
				handle_pellet_consumption(&player_x,&player_y,&player_health,pellets_x,pellets_y,pellets_type,&num_pellets_consumed, &monster_speed, &monster_color);
				handle_monster_attack(player_x,player_y,&player_health,monster_x,monster_y,&gameOver,&gameIsPaused);
				draw_game(player_x,player_y,player_health,player_color,monster_x,monster_y,
					monster_color,pellets_x,pellets_y,pellets_type,num_pellets_consumed,gameOver);
			}
			break;
		default:
			break;
		}

	}
	
	/* The user has quit.  Close the screen and end the program */
	GP142_close();

	return 0;
}


/*****************************************************************
 * FUNCTION DEFINITIONS
 *****************************************************************/


/* DRAW FUNCTIONS
 * These functions draw the different objects.  This functions don't change
 * the objects, they simply draw based on parameters calculated elsewhere.
 */


 /* draw the player based on his current position and color */
void draw_player(int player_x, int player_y, int player_color) {
	GP142_circleXY(player_color,player_x,player_y,PLAYER_WIDTH/2);
	GP142_circleXY(BLACK, player_x - (PLAYER_WIDTH/6), player_y + (PLAYER_WIDTH/6), PLAYER_WIDTH/8);
	GP142_circleXY(BLACK, player_x + (PLAYER_WIDTH/6), player_y + (PLAYER_WIDTH/6), PLAYER_WIDTH/8);
	GP142_lineXY(BLACK,player_x-(PLAYER_WIDTH/4),player_y-(PLAYER_WIDTH/5),player_x+(PLAYER_WIDTH/4),player_y-(PLAYER_WIDTH/5),PLAYER_WIDTH/12);
}


 /* draw the Evil Monster based on his current position and color */
void draw_monster(int monster_x, int monster_y, int monster_color) {
	GP142_circleXY(monster_color,monster_x,monster_y,PLAYER_WIDTH/2);
	GP142_circleXY(WHITE, monster_x, monster_y + (PLAYER_WIDTH/5), PLAYER_WIDTH/8);	
	GP142_circleXY(BLACK, monster_x, monster_y + (PLAYER_WIDTH/5)-1, PLAYER_WIDTH/20);	
	GP142_lineXY(BLACK,monster_x-(PLAYER_WIDTH/4),monster_y-(PLAYER_WIDTH/5),monster_x+(PLAYER_WIDTH/4),monster_y-(PLAYER_WIDTH/5),PLAYER_WIDTH/12);
}


/* draw a pellet based on its position and type.  We use the type to determine
   the colors of the pellet. */
void draw_pellet(int x, int y, int pellet_type) {
	int inner_color, outer_color;
	pellet_color(&inner_color, &outer_color,pellet_type);  /* get the colors of the pellet of this type */
	
	GP142_circleXY(outer_color,x,y,(int)(PELLET_WIDTH/2));
	GP142_circleXY(inner_color,x,y,(int)(PELLET_WIDTH/2*3/5));
}


/* draw the buttons at the top of the screen as well as the line across the top */
void draw_buttons(void) {
	int i;
	GP142_lineXY(BLACK,-GP142_XMAX,BUTTON_BOTTOM,GP142_XMAX,BUTTON_BOTTOM,2);
	for (i=0; i<2; i++) {
		GP142_rectangleXY(BLACK,-GP142_XMAX+5+i*BUTTON_WIDTH,BUTTON_BOTTOM,-GP142_XMAX+5+BUTTON_WIDTH+i*BUTTON_WIDTH,BUTTON_BOTTOM+BUTTON_HEIGHT-5,2);
	}
	GP142_printfXY(BLUE,-GP142_XMAX+15,BUTTON_BOTTOM+2,POINT_SIZE,"Change Color");
	GP142_printfXY(BLUE,-GP142_XMAX+25+BUTTON_WIDTH,BUTTON_BOTTOM+2,POINT_SIZE,"Right Button");
}


/* draw_game is the main drawing function.  This function takes almost all of the
   parameters in the game, as it passes them on to other functions.  This function
   makes calls to the other draw functions -- thus it is the only draw function
   called in main */

void draw_game(int player_x, int player_y, int player_health, int player_color, int monster_x, int monster_y, int monster_color, int pellets_x[],
			   int pellets_y[], int pellets_type[], int num_pellets_consumed, int gameOver) {
	int i;

	GP142_clear(); /* clear the display so we can redraw it */

	/* draw the different objects */
	draw_buttons();
	draw_player(player_x,player_y,player_color);
	draw_monster(monster_x,monster_y,monster_color);
	
	/* draw each of the different pellets -- we only draw pellets which
	   are not of NON_PELLET type (e.g. real pellets) */
	for (i=0; i < NUM_PELLET; i++) {
		if (pellets_type[i] != NON_PELLET) 
			draw_pellet(pellets_x[i],pellets_y[i],pellets_type[i]);
		
	}

	/* draw the Health and Point scores */
	GP142_printfXY(FOREST_GREEN,GP142_XMAX-84,BUTTON_BOTTOM+2,POINT_SIZE," Points %5d", num_pellets_consumed);
	GP142_printfXY(ORANGE,-GP142_XMAX + 3*BUTTON_WIDTH+60,BUTTON_BOTTOM+2,POINT_SIZE," Health %d",player_health);
	
	/* if the game is over, display so */
	if (gameOver)
		GP142_printfXY(RED,-150,0,48,"GAME OVER");
}



/*****************************************************************
 * UPDATE FUNCTIONS
 *****************************************************************/

/* these functions change how the things on the screen move */


/* this function changes the players x and y coordinates based on his current
   speed in each of those directions.  It also ensures that the player does
   not go out of the screen boundaries */
void update_player_position(int *player_x,int *player_y,int player_x_speed,int player_y_speed) {
	*player_x = *player_x + player_x_speed;
	*player_y = *player_y + player_y_speed;
	enforce_bounds(player_x,player_y);
}




/*****************************************************************
 * EVENT FUNCTIONS
 *****************************************************************/


/* these functions handle the different GP142 events */


/* this function is called if a key is pressed by the user
   the key presed, whether or not the game is paused and the
   player's current speeds are passed.  the game is paused or
   unpaused if the 'p' key is pressed.  The i,j,k,l keys change
   speed of the player in the different directions by a constant amount. */
void handle_keyboard_event(char key_pressed, int *gameIsPaused, int *player_x_speed, int *player_y_speed) {
	int new_x = *player_x_speed;  /* the original speeds */
	int new_y = *player_y_speed;
	
	/* determine which key was pressed */
	switch (key_pressed) {
	case 'i':
		new_y = new_y + PLAYER_SPEED_CHANGE;
		break;
	case 'j':
		new_x = new_x - PLAYER_SPEED_CHANGE;
		break;
	case 'k':
		new_y = new_y - PLAYER_SPEED_CHANGE;
		break;
	case 'l':
		new_x = new_x + PLAYER_SPEED_CHANGE;
		break;
	case 'p':
		*gameIsPaused = !*gameIsPaused; /* pause or unpause the game */
		/* display the word Paused if the game is paused */
		GP142_printfXY(RED,-GP142_XMAX + 3*BUTTON_WIDTH+10,BUTTON_BOTTOM+2,POINT_SIZE,"Paused");
		break;
	default:
		break;
	}

	/* change the player's speed to that of the newly calculated values */
	*player_x_speed = new_x;
	*player_y_speed = new_y;
}



/* determine which button was clicked on by the mouse */
int find_mouse_event(int mouse_x, int mouse_y) {
	if (mouse_y < BUTTON_BOTTOM)
		return NON_BUTTON; /* a button wasn't pressed */

	if (mouse_x < (-GP142_XMAX + BUTTON_WIDTH + 5))
		return LEFT_BUTTON;

	if (mouse_x < (-GP142_XMAX + 3 * BUTTON_WIDTH + 5))
		return RIGHT_BUTTON;

	return RIGHT_BUTTON;
}


/* each button has a different behavior.  given the button that was pressed,
   this function determines what to do */

void handle_mouse_event(int event_type, int *player_x, int *player_y, int *player_color,
						int *player_x_speed, int *player_y_speed, int *player_health,
						int *num_pellets_consumed, int *monster_x, int *monster_y,
						int *monster_speed, int *monster_color, int pellets_x[], int pellets_y[], int pellets_type[],
						int *gameIsPaused, int *gameOver) {

	/* figure out what to do based on which button was pressed */
	switch(event_type) {
	case LEFT_BUTTON:
		/* this button simply cycles through the different colors the player can be */
		*player_color = (*player_color + 1) % MAX_COLORS + 1;
		/* we draw the player so that the change is visible immediately */
		draw_player(*player_x,*player_y,*player_color);
		break;
	default:
		break;
	}
}

/* check whether the monster is touching the player.  If so, the player loses health points */
/* the game ends when the player's health reaches zero */
void handle_monster_attack(int player_x, int player_y, int *player_health, int monster_x,
						   int monster_y, int *gameOver, int *gameIsPaused) {
	/* the monsters are touching if they are within PLAYER_WIDTH of each other, as
	   their radii are both half PLAYER_WIDTH */
	if (distance(player_x,player_y,monster_x,monster_y) < PLAYER_WIDTH)
		*player_health = *player_health - 1; /* player lose 1 health point for each attack */
	if (*player_health == 0) {
		*gameOver = TRUE;
		*gameIsPaused = TRUE;
	}	
}

/* check to see if the player is touching any of the pellets.  If the player touches a
   pellet the pellet goes away after it produces some sort of reaction.  A pellet goes
   away if its type becomes NON_PELLET.  The player's score increases if it eats a
   pellet. Different pellets increase the score by different amounts.  Pellets can
   affect any characteristic of the game, from score, to health, to position, to anything.
   When the last pellet on the screen has been eaten, a new set of pellets are made */
void handle_pellet_consumption(int *player_x, int *player_y, int *player_health, int pellets_x[], int pellets_y[], int pellets_type[],
							   int *num_pellets_consumed, int *monster_speed, int *monster_color) {
	int i=0;
	int goodPellet = FALSE; /* used to determine if the last pellet has been eaten */
	int pellet_value = 1;   /* default pellet value */
	int pellet_type;

	/* check each pellet */
	for (i=0; i<NUM_PELLET; i++) {
		/* is it still on the screen */
		if (pellets_type[i] != NON_PELLET)	{
			/* is it being touched by the player. */
			if (distance(*player_x,*player_y,pellets_x[i],pellets_y[i]) < ((PLAYER_WIDTH + PELLET_WIDTH)/2)) {
				pellet_type = pellets_type[i]; /* keep the type aside */
				pellets_type[i] = NON_PELLET;  /* mark the pellet as eaten */
				/* now we make the pellet do whatever it is supposed to, based on its type */
				pellets_reaction(player_x,player_y,player_health,pellets_x,pellets_y,pellets_type,pellet_type,
					&pellet_value, monster_speed, monster_color);
				/* increase the score of the pellet -- it's value is determine by its reaction */
				*num_pellets_consumed = *num_pellets_consumed + pellet_value;
				
			} else {
				goodPellet = TRUE; /* this pellet remains on the screen -- it was not eaten */
			}
		}	
	}

	/* there are no more pellets -- generate some new ones */
	if (!goodPellet) {
		initialize_pellets(pellets_x,pellets_y,pellets_type);
	}
}



/* this function determines the behavior of each pellet based on its type. */

void pellets_reaction(int *player_x, int *player_y, int *player_health, int pellets_x[], int pellets_y[],
					  int pellets_type[], int pellet_type, int *pellet_value,
					  int *monster_speed, int *monster_color) {
	*pellet_value = PELLET_VALUE; /* default value of a pellet -- unless otherwise noted, 
								  eating a pellet increases the player's score by this much */

	switch(pellet_type) {
	case POINT_PELLET:
		*pellet_value = PELLET_VALUE;  /* this pellet just increases the players score by PELLET_VALUE */
		break;
	case BONUS_PELLET:      /* this pellet is worth twice a normal pellet */
		*pellet_value = 2 * PELLET_VALUE;
		break;
	case CLEARALL_PELLET:   /* this pellet clears the rest of the pellets from the screen and
							   makes all new pellets */
		initialize_pellets(pellets_x,pellets_y,pellets_type);
		break;
	case HEALTH_PELLET:     /* this pellet increase the health of the player */
		*player_health = *player_health + HEALTH_INCREASE;  /* this is the amount the player's health increases */
		break;
	default:
		break;
	}

}


/* this function is called by draw_pellet -- given a pellet_type it assigns
   the inner and outer colors of the pellet of that type */

void pellet_color(int *inner_color, int *outer_color, int pellet_type) {
	switch (pellet_type) {
	case POINT_PELLET:
		*inner_color = PEACH;
		*outer_color = SEA_GREEN;
		break;
	case BONUS_PELLET:
		*inner_color = ORANGE;
		*outer_color = MAGENTA;
		break;
	case CLEARALL_PELLET:
		*inner_color = WHITE;
		*outer_color = RED;
		break;
	case HEALTH_PELLET:
		*inner_color = OLIVE;
		*outer_color = MED_GRAY;
		break;
    default:
		*inner_color = PEACH;
		*outer_color = SEA_GREEN;
		break;
	}
}


/* this function generates a new set of NUM_PELLET pellets.  Each pellet
   is randomly given a new position and a new type */

void initialize_pellets(int pellets_x[], int pellets_y[], int pellets_type[]) {
	int i;
	int pellet_x,pellet_y,pellet_type;

	for (i=0; i<NUM_PELLET;i++) {
		new_position(&pellet_x,&pellet_y); /* random position for pellet */
		pellet_type = new_pellet_type();   /* random type */
				
		pellets_x[i]    = pellet_x;
		pellets_y[i]    = pellet_y;
		pellets_type[i] = pellet_type;
	}
}


/* returns a randomly chosen pellet type 
 * the pellet type distribution is non-uniform, that is each pellet occurs with
 * a different probablity.  choice is a number between 0 and 1.  We use it
 * to determine how often each pellet type is generated.  The difference between
 * two successive pellets values determines how the percentage of the time the later type
 * happens.  This means that:
 * pellet type    % of time it appears
 * -------------  --------------------
 * POINT_PELLET    40%
 * BONUS_PELLET    30%
 * CLEARALL_PELLET 15%
 * HEALTH_PELLET   15%
 *                total = 100%
 * When you add your own types, you should changes these values, such that your
 * pellets occur with some frequency, at least 5% of the time.
 * You do not have to keep the same percentages as given here.
 */
int new_pellet_type(void) {
	double choice = (rand()/(RAND_MAX+1.0));
			
	if (choice < .4)
		return POINT_PELLET;
	if (choice < .70)
		return BONUS_PELLET;
	if (choice < .85)
		return CLEARALL_PELLET;
	return HEALTH_PELLET;

}


/* generate a new position on the board within the approriate bounds */
void new_position(int *x, int *y) {
	*x = (int)((2.0*GP142_XMAX)*rand()/(RAND_MAX+1.0)) - GP142_XMAX;
	*y = (int)((2.0*GP142_YMAX)*rand()/(RAND_MAX+1.0)) - GP142_YMAX;	
	enforce_bounds(x,y);
}


/* make sure x and y are within the bounds defined above */
void enforce_bounds(int *x, int *y) {
	if (*x < X_LEFT_BOUND)	 *x = X_LEFT_BOUND;
	if (*x > X_RIGHT_BOUND)	 *x = X_RIGHT_BOUND;
	if (*y > Y_TOP_BOUND)	 *y = Y_TOP_BOUND;
	if (*y < Y_BOTTOM_BOUND) *y = Y_BOTTOM_BOUND;
}



/* calculate the distance between two points on the screen (x1,y1) and (x2,y2) */
int distance(int x1, int y1, int x2, int y2) {
	int x = abs(x1 - x2);
	int y = abs(y1 - y2);
	return (int)(sqrt((x*x) + (y*y)));
}



















